package com.vcarecity.elastic.publish.service.impl;

import com.fasterxml.jackson.databind.ObjectMapper;
import com.vcarecity.elastic.publish.pojo.dto.UnitAgencyDTO;
import com.vcarecity.elastic.publish.service.UnitService;
import com.vcarecity.elastic.util.AgencyPathUtil;
import lombok.extern.slf4j.Slf4j;
import org.elasticsearch.action.search.SearchRequest;
import org.elasticsearch.action.search.SearchResponse;
import org.elasticsearch.client.RequestOptions;
import org.elasticsearch.client.RestHighLevelClient;
import org.elasticsearch.index.query.BoolQueryBuilder;
import org.elasticsearch.index.query.QueryBuilder;
import org.elasticsearch.index.query.QueryBuilders;
import org.elasticsearch.index.query.TermQueryBuilder;
import org.elasticsearch.search.SearchHit;
import org.elasticsearch.search.SearchHits;
import org.elasticsearch.search.builder.SearchSourceBuilder;
import org.springframework.stereotype.Service;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Map;

/**
 * @author Kerry on 19/12/11
 */

@Slf4j
@Service
public class UnitServiceImpl implements UnitService {

    private final RestHighLevelClient restHighLevelClient;
    private final ObjectMapper objectMapper;

    public UnitServiceImpl(RestHighLevelClient restHighLevelClient,
                           ObjectMapper objectMapper) {
        this.restHighLevelClient = restHighLevelClient;
        this.objectMapper = objectMapper;
    }

    @Override
    public List<UnitAgencyDTO> getUnitAgencyByQuery(String agencyPath, String inputKeyword, int page, int size, String... indices) throws IOException {
        List<Integer> agencyIdList = AgencyPathUtil.splitAgencyId(agencyPath);
        if (agencyIdList.isEmpty()) {
            return Collections.emptyList();
        }
        QueryBuilder agencyIdBoolQuery = agencyIdListCondition(agencyIdList);
        QueryBuilder keywordBoolQuery = multiInputParamKeywordBool(inputKeyword);

        BoolQueryBuilder queryBuilder = QueryBuilders.boolQuery().must(agencyIdBoolQuery);

        if (keywordBoolQuery != null) {
            queryBuilder.must(keywordBoolQuery);
        }

        String[] queryColumns = new String[]{"unitId", "nameOfBuilding", "agencyDetail", "agencyName"};

        SearchRequest searchRequest = new SearchRequest();
        SearchSourceBuilder searchSourceBuilder = new SearchSourceBuilder();
        searchSourceBuilder.query(queryBuilder);
        searchSourceBuilder.fetchSource(queryColumns, new String[]{});
        searchSourceBuilder.from((page - 1) * size);
        searchSourceBuilder.size(size);
        searchRequest.source(searchSourceBuilder);
        searchRequest.indices(indices);


        logger.debug("getUnitAgencyByQuery:{}", searchSourceBuilder);

        SearchResponse searchResponse = restHighLevelClient.search(searchRequest, RequestOptions.DEFAULT);

        final SearchHits hits = searchResponse.getHits();
        List<UnitAgencyDTO> res = new ArrayList<>(hits.getHits().length);
        for (SearchHit hit : hits.getHits()) {
            UnitAgencyDTO unitAgencyDTO = objectMapper.readValue(hit.getSourceAsString(), UnitAgencyDTO.class);
            res.add(unitAgencyDTO);
        }
        return res;
    }

    /**
     * agency id
     * <pre>
     * {
     *   "size" : 100,
     *   "query" : {
     *     "bool" : {
     *       "must" : [
     *         {
     *           "bool" : {
     *             "must" : [
     *               {
     *                 "term" : {
     *                   "aId0" : {
     *                     "value" : 1,
     *                     "boost" : 1.0
     *                   }
     *                 }
     *               },
     *               {
     *                 "term" : {
     *                   "aId1" : {
     *                     "value" : 3,
     *                     "boost" : 1.0
     *                   }
     *                 }
     *               }
     *             ],
     *             "adjust_pure_negative" : true,
     *             "boost" : 1.0
     *           }
     *         },
     *         {
     *           "term" : {
     *             "aId2" : {
     *               "value" : 67921,
     *               "boost" : 1.0
     *             }
     *           }
     *         }
     *       ],
     *       "adjust_pure_negative" : true,
     *       "boost" : 1.0
     *     }
     *   },
     *   "_source" : {
     *     "includes" : [
     *       "unitId"
     *     ],
     *     "excludes" : [ ]
     *   },
     *   "sort" : [
     *     {
     *       "_doc" : {
     *         "order" : "asc"
     *       }
     *     }
     *   ]
     * }
     * </pre>
     *
     * @param agencyIds
     * @return
     */
    private QueryBuilder agencyIdListCondition(List<Integer> agencyIds) {

        if (agencyIds == null || agencyIds.size() == 0) {
            return null;
        }

        boolean isOnlyOnceId = agencyIds.size() == 1;

        final BoolQueryBuilder boolQueryBuilder = QueryBuilders.boolQuery();

        final Map<String, Integer> map = AgencyPathUtil.agencyPathQueryMap(agencyIds);

        for (Map.Entry<String, Integer> entry : map.entrySet()) {
            TermQueryBuilder termQueryBuilder = QueryBuilders.termQuery(entry.getKey(), entry.getValue());
            if (isOnlyOnceId) {
                return termQueryBuilder;
            }
            boolQueryBuilder.must(termQueryBuilder);
        }
        return boolQueryBuilder;
    }


    /**
     * key word
     * <pre>
     *  {
     *   "size" : 100,
     *   "query" : {
     *     "bool" : {
     *       "must" : [
     *         {
     *           "bool" : {
     *             "must" : [
     *               {
     *                 "bool" : {
     *                   "must" : [
     *                     {
     *                       "term" : {
     *                         "aId0" : {
     *                           "value" : 1,
     *                           "boost" : 1.0
     *                         }
     *                       }
     *                     },
     *                     {
     *                       "term" : {
     *                         "aId1" : {
     *                           "value" : 3,
     *                           "boost" : 1.0
     *                         }
     *                       }
     *                     }
     *                   ],
     *                   "adjust_pure_negative" : true,
     *                   "boost" : 1.0
     *                 }
     *               },
     *               {
     *                 "term" : {
     *                   "aId2" : {
     *                     "value" : 67921,
     *                     "boost" : 1.0
     *                   }
     *                 }
     *               }
     *             ],
     *             "adjust_pure_negative" : true,
     *             "boost" : 1.0
     *           }
     *         },
     *         {
     *           "bool" : {
     *             "should" : [
     *               {
     *                 "bool" : {
     *                   "should" : [
     *                     {
     *                       "wildcard" : {
     *                         "agencyName.keyword" : {
     *                           "wildcard" : "*鱼峰区*",
     *                           "boost" : 1.0
     *                         }
     *                       }
     *                     },
     *                     {
     *                       "wildcard" : {
     *                         "agencyDetail.keyword" : {
     *                           "wildcard" : "*鱼峰区*",
     *                           "boost" : 1.0
     *                         }
     *                       }
     *                     }
     *                   ],
     *                   "adjust_pure_negative" : true,
     *                   "boost" : 1.0
     *                 }
     *               },
     *               {
     *                 "wildcard" : {
     *                   "nameOfBuilding.keyword" : {
     *                     "wildcard" : "*鱼峰区*",
     *                     "boost" : 1.0
     *                   }
     *                 }
     *               }
     *             ],
     *             "adjust_pure_negative" : true,
     *             "boost" : 1.0
     *           }
     *         }
     *       ],
     *       "adjust_pure_negative" : true,
     *       "boost" : 1.0
     *     }
     *   },
     *   "_source" : {
     *     "includes" : [
     *       "unitId"
     *     ],
     *     "excludes" : [ ]
     *   },
     *   "sort" : [
     *     {
     *       "_doc" : {
     *         "order" : "asc"
     *       }
     *     }
     *   ]
     * }
     * </pre>
     *
     * @param keyword
     * @return
     */
    private BoolQueryBuilder multiInputParamKeywordBool(String keyword) {
        if (keyword == null || keyword.trim().length() == 0) {
            return null;
        }

        final String[] keyNames = new String[]{
                "agencyName.keyword",
                "agencyDetail.keyword",
                "nameOfBuilding.keyword",
        };

        final BoolQueryBuilder boolQueryBuilder = QueryBuilders.boolQuery();

        for (String kn : keyNames) {
            QueryBuilder queryBuilder = QueryBuilders.wildcardQuery(kn, "*" + keyword + "*");
            // QueryBuilder queryBuilder = QueryBuilders.matchQuery(kn, keyword);
            boolQueryBuilder.should(queryBuilder);
        }
        return boolQueryBuilder;
    }

}
